import 'dart:async';
import 'dart:convert';

import 'package:meta/meta.dart';
import 'package:web_socket_channel/web_socket_channel.dart';

import 'model.dart';
import 'serializer.dart';

class KeenApiClient<Action, Update> {
  KeenApiClient.fromWebSocket({
    @required WebSocketChannel channel,
    @required KeenSerializer<Action> actionSerializer,
    @required KeenDeserializer<Update> updateDeserializer,
  })  : _channel = channel,
        _actionSerializer = actionSerializer,
        _updateDeserializer = updateDeserializer {
    _subscription = _channel.stream.listen(
      _onMessage,
      onDone: () async {
        _isDisposed = true;
        await _subscription.cancel();
        await _channel.sink.close(2000);
        await _controller.close();
      },
    );
  }

  final WebSocketChannel _channel;
  final KeenSerializer<Action> _actionSerializer;
  final KeenDeserializer<Update> _updateDeserializer;

  static int _nextRequestId = 0;
  bool _isDisposed = false;

  bool get isDisposed => _isDisposed;

  StreamSubscription<dynamic> _subscription;

  final StreamController<Update> _controller =
      StreamController<Update>.broadcast();

  final Map<String, Completer<void>> _completers = {};

  Stream<Update> get updates => _controller.stream;

  void _onMessage(dynamic message) {
    final output = KeenResponse.fromJson(jsonDecode(message));
    output.when(
      update: (requestId, value) {
        if (requestId != null) {
          _completers[requestId].complete();
          _completers.remove(requestId);
        }
        _controller.sink.add(_updateDeserializer(value));
      },
      empty: (requestId) {
        _completers[output.requestId].complete();
        _completers.remove(output.requestId);
      },
      error: (requestId, error) {
        _completers[requestId].completeError(error);
        _completers.remove(requestId);
      },
    );
  }

  Future<void> action(Action action) {
    if (_isDisposed) throw Exception('Client disposed, create a new one');
    final input = KeenRequest(
      requestId: (_nextRequestId++).toString(),
      value: _actionSerializer(action),
    );
    final completer = Completer<void>();
    _completers[input.requestId] = completer;
    _channel.sink.add(jsonEncode(input.toJson()));
    return completer.future;
  }

  Future<void> close() async {
    if (!_isDisposed) {
      _isDisposed = true;
      await _subscription.cancel();
      await _channel.sink.close(2000);
      await _controller.close();
    }
  }
}
