import 'dart:convert';
import 'dart:io';

import 'package:aes_crypt/aes_crypt.dart';
import 'package:meta/meta.dart';
import 'package:http/http.dart' as http;
import 'package:web_socket_channel/io.dart';
import 'package:web_socket_channel/web_socket_channel.dart';

import 'encryption.dart';
import 'model.dart';

/// A basic authentication service based on one-time passwords sent
/// by email.
class KeenAuthClient {
  final String url;
  const KeenAuthClient({
    @required this.url,
  });

  Future<SharedSecret> sendOneTimePasswordByEmail({
    @required String email,
    @required String deviceName,
    String locale = 'en_US',
  }) async {
    final encryption = PasswordEncryption.generate();
    final result = await http.post(
      '$url/auth/send',
      body: jsonEncode({
        'email': email,
        'key': encryption.key,
        'iv': encryption.iv,
        'deviceName': deviceName,
        'locale': locale,
      }),
    );
    result.throwIfInvalid();
    final json = jsonDecode(result.body);
    return SharedSecret(
      key: encryption.key,
      iv: encryption.iv,
      expiry: DateTime.parse(json['expiry']),
    );
  }

  Future<Token> getToken({
    @required SharedSecret sharedSecret,
    @required String oneTimePassword,
  }) async {
    final encryption = PasswordEncryption(
      key: sharedSecret.key,
      iv: sharedSecret.iv,
    );
    //
    final result = await http.post(
      '$url/auth/token',
      body: jsonEncode(
        {
          'otp': encryption.decryptPassword(oneTimePassword),
        },
      ),
    );
    result.throwIfInvalid();
    return Token.fromJson(jsonDecode(result.body));
  }

  Future<Token> refreshToken(Token token) async {
    final result = await http.post(
      '$url/auth/refresh',
      headers: token.asHeader(),
      body: jsonEncode(
        {
          'token': token.refreshToken,
        },
      ),
    );
    result.throwIfInvalid();
    return Token.fromJson(jsonDecode(result.body));
  }

  WebSocketChannel createWebSocket(
    Token token, {
    Map<String, String> headers,
  }) {
    if (token.isExpired) {
      throw Exception('Token is expired');
    }
    final wsUrl =
        url.replaceAll('http://', 'ws://').replaceAll('https://', 'wss://');
    return IOWebSocketChannel.connect(
      '$wsUrl/api',
      headers: {
        'Authorization': token.accessToken.encode(),
        if (headers != null) ...headers,
      },
    );
  }
}

extension ResponseExtensions on http.Response {
  void throwIfInvalid() {
    if (statusCode < 200 || statusCode >= 300) {
      throw KeenHttpException(statusCode, body);
    }
  }
}

class KeenHttpException {
  final int statusCode;
  final String message;
  const KeenHttpException(this.statusCode, this.message);
}
