import 'dart:convert';

import 'package:meta/meta.dart';
import 'package:freezed_annotation/freezed_annotation.dart';

part 'model.freezed.dart';
part 'model.g.dart';

/// An authentication token required by requests to the api.
@freezed
abstract class SharedSecret with _$SharedSecret {
  const factory SharedSecret({
    /// The key that is used to decrypt the upcoming token from the email link.
    @required String key,

    /// The IV vector for encryption.
    @required String iv,

    /// The maximum date after which the link key isn't valid anymore.
    @required DateTime expiry,
  }) = _SharedSecret;

  factory SharedSecret.fromJson(Map<String, dynamic> json) =>
      _$SharedSecretFromJson(json);
}

/// An authentication token required by requests to the api.
@freezed
abstract class Token with _$Token {
  const factory Token({
    /// The access token must be joined to all request to the api as an
    /// `Authorization` header.
    @required JwtToken accessToken,

    /// The [refreshToken] can be used to acquire a new token from the api.
    @required String refreshToken,
  }) = _Token;

  factory Token.fromJson(Map<String, dynamic> json) => _$TokenFromJson(json);
}

extension TokenExtension on Token {
  bool get isExpired => accessToken.payload.exp.isBefore(DateTime.now());
  Map<String, String> asHeader() {
    return {
      'Authorization': 'Bearer ${accessToken.encode()}',
    };
  }
}

@freezed
abstract class JwtToken with _$JwtToken {
  const factory JwtToken({
    /// The serialized value of the token
    @required String signature,
    @required JwtHeader header,
    @required JwtPayload payload,
  }) = _JwtToken;

  factory JwtToken.fromJson(Map<String, dynamic> json) =>
      _$JwtTokenFromJson(json);

  static JwtToken decode(String encoded) {
    final splits = encoded.split('.');
    final jsonHeader = utf8.decode(
      base64Decode(splits[0]),
    );
    final jsonPayload = utf8.decode(
      base64Decode(splits[1]),
    );
    return JwtToken(
      header: JwtHeader.fromJson(
        jsonDecode(jsonHeader),
      ),
      payload: JwtPayload.fromJson(
        jsonDecode(jsonPayload),
      ),
      signature: splits[2],
    );
  }
}

@freezed
abstract class JwtHeader with _$JwtHeader {
  const factory JwtHeader({
    @required String alg,
    @required String typ,
  }) = _JwtHeader;

  factory JwtHeader.fromJson(Map<String, dynamic> json) =>
      _$JwtHeaderFromJson(json);
}

@freezed
abstract class JwtPayload with _$JwtPayload {
  const factory JwtPayload({
    /// The user identifier.
    @required String userId,
    @required DateTime exp,
  }) = _JwtPayload;

  factory JwtPayload.fromJson(Map<String, dynamic> json) =>
      _$JwtPayloadFromJson(json);
}

extension JwtTokenExtension on JwtToken {
  String encode() {
    final jsonHeader = base64Encode(utf8.encode(jsonEncode(header.toJson())));
    final jsonPayload = base64Encode(utf8.encode(jsonEncode(payload.toJson())));
    return '$jsonHeader.$jsonPayload.$signature';
  }
}
