import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:keen/client.dart';
import 'package:meta/meta.dart';

import '../server.dart';

abstract class KeenApiServer<SessionInfo, Action, Update>
    extends KeenServerController {
  KeenApiServer({
    @required this.updateSerializer,
    @required this.actionDeserializer,
  });

  final KeenSerializer<Update> updateSerializer;
  final KeenDeserializer<Action> actionDeserializer;
  final List<KeenServerSession<SessionInfo, Action, Update>> sessions = [];

  /// Extract session informations from the [request].
  Future<SessionInfo> getSessionInfo(HttpRequest request);

  /// Create a session from the [request], with the associated [info] metadata.
  @override
  FutureOr<bool> process(
    HttpRequest request,
  ) async {
    final segments = request.uri.pathSegments;
    if (segments.length == 1 && segments[0] == 'api') {
      final info = await getSessionInfo(request);
      final socket = await WebSocketTransformer.upgrade(request);

      sessions.add(
        KeenServerSession<SessionInfo, Action, Update>(
          server: this,
          info: info,
          socket: socket,
        ),
      );

      return true;
    }
    return false;
  }

  /// Sends the [update] message to the sessions that satisfy the [predicate].
  void update(
    Update update,
    bool Function(SessionInfo info) predicate,
  ) {
    for (var session in sessions) {
      if (predicate(session.info)) {
        session.update(update);
      }
    }
  }

  /// Whenever an [action] is sent to a [session].
  ///
  /// An update can be returned back to the client. If null, then
  /// a [KeenEmptyResponse] is returned instead.
  FutureOr<Update> onAction(
    SessionInfo session,
    Action action,
  );
}

/// An active user session.
class KeenServerSession<SessionInfo, Action, Update> {
  final KeenApiServer<SessionInfo, Action, Update> server;
  final SessionInfo info;
  final WebSocket socket;
  KeenServerSession({
    @required this.server,
    @required this.info,
    @required this.socket,
  }) {
    socket.listen(
      _onMessage,
      onDone: () {
        server.sessions.remove(this);
      },
    );
  }

  /// Sends the [update] message to the client.
  void update(Update update) {
    final response = KeenResponse.update(
      value: server.updateSerializer(update),
    );
    socket.add(jsonEncode(response.toJson()));
  }

  Future<void> _onMessage(dynamic event) async {
    final request = KeenRequest.fromJson(jsonDecode(event));

    final action = server.actionDeserializer(request.value);
    KeenResponse response;
    try {
      final update = await server.onAction(info, action);
      if (update == null) {
        response = KeenResponse.empty(requestId: request.requestId);
      } else {
        response = KeenResponse.update(
          requestId: request.requestId,
          value: server.updateSerializer(update),
        );
      }
    } on KeenError catch (e) {
      response = KeenResponse.error(
        requestId: request.requestId,
        error: e,
      );
    } catch (e) {
      response = KeenResponse.error(
        requestId: request.requestId,
        error: KeenError(
          code: KeenErrorCodes.internalServerError,
          message: 'Internal server error',
        ),
      );
    }

    socket.add(jsonEncode(response.toJson()));
  }

  Future<void> close() async {
    server.sessions.remove(this);
    await socket.close();
  }
}
