import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:aes_crypt/aes_crypt.dart';
import 'package:keen/client.dart';
import 'package:keen/src/client/auth/encryption.dart';
import 'package:keen/src/server/auth/users/storage/storage.dart';
import 'package:meta/meta.dart';
import 'package:uuid/uuid.dart';

import '../server.dart';
import 'otp/emails/emailer.dart';
import 'jwt/jwt.dart';
import 'otp/storage/storage.dart';
import 'users/user.dart';

var _uuid = Uuid();

class KeenAuthServer extends KeenServerController {
  final KeenEmailer emailer;
  final Duration oneTimePasswordExpiry;
  final Duration tokenExpiry;
  final UserStorage userStorage;
  final OtpStorage otpStorage;
  final JwtManager jwt;

  KeenAuthServer({
    @required String signKey,
    @required this.emailer,
    @required this.userStorage,
    this.oneTimePasswordExpiry = const Duration(minutes: 5),
    this.tokenExpiry = const Duration(hours: 4),
    OtpStorage otpStorage,
  })  : otpStorage = otpStorage ?? MemoryOtpStorage(),
        jwt = JwtManager(
          key: signKey,
          expiry: tokenExpiry,
        );

  @override
  FutureOr<bool> process(HttpRequest request) async {
    final segments = request.uri.pathSegments;
    if (request.method == 'POST' &&
        segments.length == 2 &&
        segments[0] == 'auth') {
      final bodyContent = await utf8.decodeStream(request);
      final bodyJson = bodyContent.isNotEmpty ? jsonDecode(bodyContent) : null;
      if (segments[1] == 'send') {
        await _send(request.response, bodyJson);
        return true;
      } else if (segments[1] == 'token') {
        await _token(request.response, bodyJson);
        return true;
      } else if (segments[1] == 'refresh') {
        await _refresh(request, bodyJson);
        return true;
      }
    }

    return false;
  }

  /// Send an email with a one time password to the email.
  Future<void> _send(
    HttpResponse response,
    Map<String, dynamic> body,
  ) async {
    final email = body['email'];
    // TODO validate email

    final locale = body.containsKey('locale') ? body['locale'] : 'en_US';
    final deviceName = body['deviceName'];
    final key = body['key'];
    final iv = body['iv'];

    final otpRequest = await otpStorage.create(
      email: email,
      expiry: DateTime.now().add(oneTimePasswordExpiry),
      locale: locale,
      deviceName: deviceName,
      encrypt: PasswordEncryption(key: key, iv: iv).encryptPassword,
    );

    await emailer(otpRequest);
    response.statusCode = HttpStatus.ok;
    response.write(
      jsonEncode(
        {
          'expiry': otpRequest.expiry.toIso8601String(),
        },
      ),
    );
    await response.close();
  }

  Future<void> _token(
    HttpResponse response,
    Map<String, dynamic> body,
  ) async {
    final String otp = body['otp'];
    final request = await otpStorage.find(otp);
    if (request.isNotExpired) {
      await otpStorage.delete(request);
      final user = await userStorage.findOrCreate(request.email);
      final token = Token(
        accessToken: jwt.create(user.id),
        refreshToken: _uuid.v4(),
      );
      final updatedUser = User(
        id: user.id,
        email: user.email,
        creationDate: user.creationDate,
        refreshTokens: [
          ...user.refreshTokens,
          token.refreshToken,
        ],
      );
      await userStorage.save(updatedUser);
      response.statusCode = HttpStatus.ok;
      response.write(jsonEncode(token.toJson()));
    } else {
      response.statusCode = HttpStatus.gone;
    }
    await response.close();
  }

  Future<void> _refresh(
    HttpRequest request,
    Map<String, dynamic> body,
  ) async {
    final String refreshToken = body['token'];
    final previousToken = jwt.fromHeaders(request);

    final user = await userStorage.getUser(previousToken.payload.userId);
    if (user.refreshTokens.contains(refreshToken)) {
      /// Creating a fresh token and removing old refresh token
      final newToken = Token(
        accessToken: jwt.create(user.id),
        refreshToken: _uuid.v4(),
      );
      final updatedUser = User(
        id: user.id,
        email: user.email,
        creationDate: user.creationDate,
        refreshTokens: [
          ...user.refreshTokens.where((x) => x != refreshToken),
          newToken.refreshToken,
        ],
      );
      await userStorage.save(updatedUser);

      request.response.statusCode = HttpStatus.ok;
      request.response.write(jsonEncode(newToken.toJson()));
    } else {
      request.response.statusCode = HttpStatus.notFound;
    }
    await request.response.close();
  }
}
