import 'dart:convert';
import 'dart:io';

import 'package:crypto/crypto.dart';
import 'package:keen/client.dart';
import 'package:meta/meta.dart';

/// Allow to create and verify tokens regarding a signing algorithm.
class JwtManager {
  /// Create a new manager with the given signing [key].
  ///
  /// All created tokens will have an expiry of [expiry].
  JwtManager({
    @required String key,
    @required this.expiry,
  }) : key = utf8.encode(key);

  /// The signing key used to encrypt the signature part of tokens.
  final List<int> key;

  /// The period after created tokens become invalid.
  final Duration expiry;

  /// Create a new token with the [userId].
  JwtToken create(String userId) {
    final payload = JwtPayload(
      userId: userId,
      exp: DateTime.now().add(expiry),
    );

    final header = JwtHeader(
      alg: 'SHA256',
      typ: 'JWT',
    );

    return JwtToken(
      signature: _createSignature(header, payload),
      header: header,
      payload: payload,
    );
  }

  /// Read the token from the [request] headers.
  JwtToken fromHeaders(HttpRequest request) {
    final result =
        request.headers['Authorization'].first.replaceAll('Bearer ', '');
    return JwtToken.decode(result);
  }

  /// Verify that the signature part of the [token] is valid with the
  /// current signing [key].
  bool verify(JwtToken token) {
    final verifiedSignature = _createSignature(token.header, token.payload);
    return verifiedSignature == token.signature;
  }

  /// Encodes the header and payload and then encrypt them with the current
  /// signing [key].
  String _createSignature(JwtHeader header, JwtPayload payload) {
    final headerB64 = base64Encode(utf8.encode(jsonEncode(header.toJson())));
    final payloadB64 = base64Encode(utf8.encode(jsonEncode(payload.toJson())));
    var bytes = utf8.encode('$headerB64.$payloadB64');

    var hmacSha256 = Hmac(sha256, key);
    return base64Encode(hmacSha256.convert(bytes).bytes);
  }
}
