import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:meta/meta.dart';

part 'otp.freezed.dart';
part 'otp.g.dart';

/// An authentication request for a user with [email].
///
/// The user should send the [oneTimePassword] (that is sent by email)
/// to get a unique token idenfying him.
///
/// The request becomes invalid after [expiry].
@freezed
abstract class OtpRequest with _$OtpRequest {
  const factory OtpRequest({
    /// The email of the user that requested authentication.
    @required String email,

    /// The device name from which the password has been requested.
    @required String deviceName,

    /// The user's locale code.
    @required String locale,

    /// The expiry date after which the request becomes invalid.
    @required DateTime expiry,

    /// The password sent to the user to validate its identity.
    @required String oneTimePassword,

    /// The password sent to the user to validate its identity.
    @required String encryptedOneTimePassword,
  }) = _OtpRequest;

  /// Create a request from a [json] map.
  factory OtpRequest.fromJson(Map<String, dynamic> json) =>
      _$OtpRequestFromJson(json);
}

/// Various helpers on top of the request type.
///
/// See also:
/// * [OtpRequest] the extended type.
extension OtpRequestExtension on OtpRequest {
  /// Indicate whether the token is expired (i.e [expiry] is before now).
  bool get isExpired => expiry.isBefore(DateTime.now());

  /// Indicate whether the token is not expired (i.e [expiry] is after or same as now).
  bool get isNotExpired => !isExpired;
}
