import 'package:meta/meta.dart';
import 'package:pedantic/pedantic.dart';
import 'package:uuid/uuid.dart';

import '../otp.dart';

export '../otp.dart';

var _uuid = Uuid();

/// A storage for all pending authentification requests from users.
///
/// See also:
/// * [MemoryOtpStorage] that stores requests in memory.
abstract class OtpStorage {
  const OtpStorage();

  /// Creates a new password request for the given [email].
  ///
  /// The request should expire after [expiry] date and be deleted
  /// from the storage in such a case.
  Future<OtpRequest> create({
    @required String email,
    @required String locale,
    @required String deviceName,
    @required DateTime expiry,
    @required String Function(String password) encrypt,
  });

  /// Find the request associated to the [oneTimePassword].
  Future<OtpRequest> find(String oneTimePassword);

  /// Delete a [request] from the storage.
  Future<void> delete(OtpRequest request);
}

/// An authentication request storage that only keeps data in memory.
///
/// If the server is restarted, then all pending requests are lost and
/// they become invalid for users.
class MemoryOtpStorage extends OtpStorage {
  MemoryOtpStorage() : _requests = <OtpRequest>[];

  final List<OtpRequest> _requests;

  @override
  Future<OtpRequest> find(String oneTimePassword) {
    return Future.value(
      _requests.firstWhere(
        (x) => x.oneTimePassword == oneTimePassword,
        orElse: () => null,
      ),
    );
  }

  @override
  Future<void> delete(OtpRequest request) {
    _requests.removeWhere((x) => x.oneTimePassword == request.oneTimePassword);
    return Future.value();
  }

  @override
  Future<OtpRequest> create({
    @required String email,
    @required String locale,
    @required String deviceName,
    @required DateTime expiry,
    @required String Function(String password) encrypt,
  }) {
    final password = _uuid.v4();
    final result = OtpRequest(
      email: email,
      expiry: expiry,
      deviceName: deviceName,
      locale: locale,
      oneTimePassword: password,
      encryptedOneTimePassword: encrypt(password),
    );
    _requests.add(result);
    unawaited(_autoCleanAfterExpiry(result));
    return Future.value(result);
  }

  /// Removes [request] from [_requests] after its expiry.
  Future<void> _autoCleanAfterExpiry(OtpRequest request) async {
    final duration = DateTime.now().difference(request.expiry);
    if (!duration.isNegative) {
      await Future.delayed(duration);
    }
    _requests.removeWhere((element) => false);
  }
}
