import 'dart:convert';
import 'dart:io';

import 'storage.dart';

/// Identical as the [MemoryUserStorage] but persists the list of users
/// in a local [file].
///
/// This storage should be used carefully, and only if only a few users (several hundreds) are
/// registered. Any database storage may be more appropriated if there's more users.
class FileUserStorage extends MemoryUserStorage {
  /// Create a new storage where users are persisted in [file] and in memory.
  FileUserStorage(this.file);

  /// The file where users are persisted.
  final File file;

  /// Whether the file content has already been loaded.
  bool _isLoaded = false;

  Future<void> _save() {
    final json = users.map((x) => x.toJson()).toList();
    return file.writeAsString(jsonEncode(json));
  }

  @override
  Future<User> findOrCreate(String email) async {
    await _load();
    return super.findOrCreate(email);
  }

  @override
  Future<User> getUser(String id) async {
    await _load();
    return super.getUser(id);
  }

  @override
  Future<void> save(User user) async {
    await _save();
    return super.save(user);
  }

  Future<void> _load() async {
    if (!_isLoaded) {
      if (await file.exists()) {
        final content = await file.readAsString();
        final List json = await jsonDecode(content);
        users = json.map((x) => User.fromJson(x)).toList();
      }
      _isLoaded = true;
    }
  }
}
