import 'package:keen/src/server/auth/otp/otp.dart';
import 'package:meta/meta.dart';
import 'package:postgres/postgres.dart';

import 'storage.dart';

class PostgresUserStorage extends UserStorage {
  const PostgresUserStorage({
    @required this.host,
    @required this.port,
    @required this.databaseName,
    this.tableName = 'user',
    this.username,
    this.password,
    this.timeoutInSeconds = 30,
    this.queryTimeoutInSeconds = 30,
    this.timeZone = 'UTC',
    this.useSSL = false,
    this.isUnixSocket = false,
  });

  final int port;
  final String host;
  final String databaseName;
  final String tableName;
  final String username;
  final String password;
  final int timeoutInSeconds;
  final int queryTimeoutInSeconds;
  final String timeZone;
  final bool useSSL;
  final bool isUnixSocket;

  Future<T> _run<T>(
    Future<T> Function(PostgreSQLConnection connection) execute,
  ) async {
    final connection = PostgreSQLConnection(
      host,
      port,
      databaseName,
      username: username,
      password: password,
      timeoutInSeconds: timeoutInSeconds,
      queryTimeoutInSeconds: queryTimeoutInSeconds,
      timeZone: timeZone,
      useSSL: useSSL,
      isUnixSocket: isUnixSocket,
    );

    final error = await connection.open();
    if (error != null) throw error;
    final result = await execute(connection);
    await connection.close();
    return result;
  }

  @override
  Future<User> findOrCreate(String email) {
    return _run((connection) async {
      final creationDate = DateTime.now();
      final result = await connection.mappedResultsQuery(
        '''
INSERT INTO $tableName (email, creationDate, refreshTokens) 
VALUES (@email, @creationDate, {} }) 
ON CONFLICT (email)
DO NOTHING
''',
        substitutionValues: {
          'email': email,
          'creationDate': creationDate,
        },
      );

      return User(
        id: result.first[tableName]['id'],
        email: email,
        creationDate: creationDate,
        refreshTokens: [],
      );
    });
  }

  @override
  Future<User> getUser(String id) async {
    return _run((connection) async {
      final results = await connection.mappedResultsQuery(
        'SELECT id, email, creationDate, refreshTokens FROM $tableName WHERE id = @id',
        substitutionValues: {
          'id': id,
        },
      );

      if (results.isEmpty) throw Exception('Not found');
      final row = results.first[tableName];
      return User.fromJson(row);
    });
  }

  @override
  Future<void> save(User user) {
    return _run((connection) async {
      await connection.mappedResultsQuery(
        '''
INSERT INTO $tableName (id, email, creationDate, refreshTokens) 
VALUES (@id, @email, @creationDate, { ${user.refreshTokens.map((x) => '"$x"').join(',')} }) 
ON CONFLICT (id)
DO 
   UPDATE SET refreshTokens = EXCLUDED.refreshTokens;
''',
        substitutionValues: user.toJson(),
      );
    });
  }
}
