export '../user.dart';

import 'package:keen/src/server/auth/users/user.dart';
import 'package:meta/meta.dart';

/// Store all users.
abstract class UserStorage {
  const UserStorage();

  /// Find the user with [email] if exists, else a new one
  /// is created in the store.
  Future<User> findOrCreate(String email);

  /// Get the user with the given [id].
  Future<User> getUser(String id);

  /// Update the [user].
  Future<void> save(User user);
}

/// A user storage that only keeps users in memory.
///
/// The main use case here is testing since users should be
/// persisted between server sessions.
class MemoryUserStorage extends UserStorage {
  MemoryUserStorage() : users = <User>[];

  @protected
  List<User> users;

  int _nextUserId = 0;

  @override
  Future<User> findOrCreate(String email) {
    return Future.value(
      users.firstWhere(
        (x) => x.email == email,
        orElse: () {
          final newUser = User(
            email: email,
            creationDate: DateTime.now(),
            id: (_nextUserId++).toString(),
            refreshTokens: [],
          );
          users.add(newUser);
          return newUser;
        },
      ),
    );
  }

  @override
  Future<User> getUser(String id) {
    return Future.value(
      users.firstWhere(
        (x) => x.id == id,
        orElse: () => throw Exception('User not found'),
      ),
    );
  }

  @override
  Future<void> save(User user) {
    users.removeWhere((x) => x.id == user.id);
    users.add(user);
    return Future.value();
  }
}
